@@ -330,22 +330,32 @@ static void obex_request_done(GwObex *ctx, obex_object_t *object,
     }
 }
 
-static void get_target_size_and_time(obex_t *handle, obex_object_t *object,
-                                     gint *size, time_t *time) {
+static void get_non_body_headers(obex_t *handle, obex_object_t *object,
+                                     struct gw_obex_xfer *xfer) {
     obex_headerdata_t hv;
     uint8_t hi;
     unsigned int hlen;
 
-    *size = GW_OBEX_UNKNOWN_LENGTH;
-    *time = -1;
+    xfer->target_size = GW_OBEX_UNKNOWN_LENGTH;
+    xfer->modtime = -1;
 
     while (OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
         switch (hi) {
             case OBEX_HDR_LENGTH:
-                *size = hv.bq4; //(gint) g_ntohl(hv.bq4);
+                xfer->target_size = hv.bq4; //(gint) g_ntohl(hv.bq4);
                 break;
             case OBEX_HDR_TIME:
-                *time = parse_iso8601((char *)hv.bs, hlen);
+                xfer->modtime = parse_iso8601((char *)hv.bs, hlen);
+                break;
+            case OBEX_HDR_APPARAM:
+                g_free(xfer->apparam_buf);
+                xfer->apparam_buf = g_try_malloc(hlen);
+		if (xfer->apparam_buf) {
+                    memcpy(xfer->apparam_buf, hv.bs, hlen);
+                    xfer->apparam_size = hlen;
+                }
+                else
+                    xfer->apparam_size = 0;
                 break;
             default:
                 break;
@@ -370,8 +380,7 @@ static void obex_readstream(GwObex *ctx, obex_object_t *object) {
     }
 
     if (ctx->xfer->counter == 0) {
-        get_target_size_and_time(ctx->handle, object,
-                                 &xfer->target_size, &xfer->modtime);
+        get_non_body_headers(ctx->handle, object, xfer);
         show_headers(ctx->handle, object);
     }
 
@@ -594,10 +603,11 @@ void obex_link_error(GwObex *ctx) {
     }
     if (ctx->xfer) {
         /* Check that buffer is owned by us */
-        if (!(ctx->obex_op == OBEX_CMD_PUT && ctx->xfer->stream_fd < 0))
+        if (!(ctx->obex_op == OBEX_CMD_PUT && ctx->xfer->stream_fd < 0)) {
             g_free(ctx->xfer->buf);
-        ctx->xfer->buf = NULL;
-        ctx->xfer->buf_size = 0;
+            ctx->xfer->buf = NULL;
+            ctx->xfer->buf_size = 0;
+	}
 	ctx->xfer->do_cb = TRUE;
     }
 }
@@ -828,6 +838,7 @@ out:
 
 gboolean gw_obex_get(GwObex *ctx,
                      const gchar *local, const gchar *remote, const gchar *type,
+                     const guint8 *apparam, gint apparam_size,
                      gchar **buf, gint *buf_size, int stream_fd,
                      gboolean async) {
     gboolean ret = FALSE;
@@ -853,6 +864,11 @@ gboolean gw_obex_get(GwObex *ctx,
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_CONNECTION, hv, 4, 0);
     }
 
+    if (apparam && apparam_size > 0) {
+        hv.bs = (unsigned char *)apparam;
+        OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_APPARAM, hv, apparam_size, 0);
+    }
+
     if (type) {
         hv.bs = (unsigned char *)type;
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_TYPE, hv, strlen(type) + 1, 0);
@@ -939,6 +955,7 @@ out:
 
 gboolean gw_obex_put(GwObex *ctx,
                      const gchar *local, const gchar *remote, const gchar *type,
+                     const guint8 *apparam, gint apparam_size,
                      const gchar *buf, gint object_size, time_t object_time,
                      int stream_fd, gboolean async) {
     gboolean ret = FALSE;
@@ -954,16 +971,15 @@ gboolean gw_obex_put(GwObex *ctx,
         return FALSE;
     }
 
-    ctx->obex_op = OBEX_CMD_PUT;
-
     if (remote) {
         uname_len = get_uname(&uname, remote);
         if (uname_len < 0) {
             ctx->error = GW_OBEX_ERROR_INVALID_PARAMS;
-            goto out;
+            return FALSE;
         }
     }
 
+    ctx->obex_op = OBEX_CMD_PUT;
     ctx->xfer = gw_obex_xfer_new(ctx, async, stream_fd);
 
     if (local) {
@@ -1015,6 +1031,11 @@ gboolean gw_obex_put(GwObex *ctx,
         OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_TYPE, hv, strlen(type) + 1, 0);
     }
 
+    if (apparam && apparam_size > 0) {
+        hv.bs = (unsigned char *)apparam;
+        OBEX_ObjectAddHeader(ctx->handle, object, OBEX_HDR_APPARAM, hv, apparam_size, 0);
+    }
+
     /* Try to figure out modification time if none was given */
     if (ctx->xfer->stream_fd >= 0) {
         struct stat stats;
