@@ -47,21 +47,13 @@
 #include "logging.h"
 #include "obex.h"
 #include "dbus.h"
+#include "mimetype.h"
+#include "service.h"
 
 /* Default MTU's */
 #define DEFAULT_RX_MTU 32767
 #define DEFAULT_TX_MTU 32767
 
-#define TARGET_SIZE 16
-
-static const guint8 FTP_TARGET[TARGET_SIZE] = {
-			0xF9, 0xEC, 0x7B, 0xC4,  0x95, 0x3C, 0x11, 0xD2,
-			0x98, 0x4E, 0x52, 0x54,  0x00, 0xDC, 0x9E, 0x09  };
-
-static const guint8 PBAP_TARGET[TARGET_SIZE] = {
-			0x79, 0x61, 0x35, 0xF0,  0xF0, 0xC5, 0x11, 0xD8,
-			0x09, 0x66, 0x08, 0x00,  0x20, 0x0C, 0x9A, 0x66  };
-
 /* Connection ID */
 static guint32 cid = 0x0000;
 
@@ -73,36 +65,19 @@ typedef struct {
 	guint16 mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
-struct obex_commands opp = {
-	.get		= opp_get,
-	.put		= opp_put,
-	.chkput		= opp_chkput,
-};
-
-struct obex_commands ftp = {
-	.get		= ftp_get,
-	.put		= ftp_put,
-	.chkput		= ftp_chkput,
-	.setpath	= ftp_setpath,
-};
-
-struct obex_commands pbap = {
-	.get		= pbap_get,
-	.setpath	= pbap_setpath,
-};
-
 static void os_reset_session(struct obex_session *os)
 {
-	if (os->fd > 0) {
-		close(os->fd);
-		os->fd = -1;
+	if (os->object) {
+		os->driver->close(os->object);
+		os->object = NULL;
 		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->current_folder) {
 			gchar *path;
 			path = g_build_filename(os->current_folder, os->name, NULL);
-			unlink(path);
+			os->driver->remove(path);
 			g_free(path);
 		}
 	}
+
 	if (os->name) {
 		g_free(os->name);
 		os->name = NULL;
@@ -115,6 +90,7 @@ static void os_reset_session(struct obex_session *os)
 		g_free(os->buf);
 		os->buf = NULL;
 	}
+	os->driver = NULL;
 	os->aborted = FALSE;
 	os->offset = 0;
 	os->size = OBJECT_SIZE_DELETE;
@@ -143,9 +119,6 @@ static void obex_session_free(struct obex_session *os)
 	if (os->io)
 		g_io_channel_unref(os->io);
 
-	if (os->target && !memcmp(os->target, PBAP_TARGET, TARGET_SIZE))
-		pbap_phonebook_context_destroy(os);
-
 	g_free(os);
 }
 
@@ -205,6 +178,8 @@ static void cmd_connect(struct obex_session *os,
 	guint hlen, newsize;
 	guint16 mtu;
 	guint8 hi;
+	const guint8 *target = NULL, *who = NULL;
+	guint target_size = 0, who_size = 0;
 
 	if (OBEX_ObjectGetNonHdrData(obj, &buffer) != sizeof(*nonhdr)) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -227,57 +202,32 @@ static void cmd_connect(struct obex_session *os,
 	os->cid = ++cid;
 
 	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
-		if (hi != OBEX_HDR_TARGET || hlen != TARGET_SIZE)
-			continue;
-
-		if (memcmp(hd.bs, FTP_TARGET, TARGET_SIZE) == 0 &&
-				os->server->services &
-						(OBEX_FTP | OBEX_PCSUITE)) {
-			os->target = FTP_TARGET;
-			os->cmds = &ftp;
+		switch (hi) {
+		case OBEX_HDR_WHO:
+			who = hd.bs;
+			who_size = hlen;
 			break;
-		}
-
-		if (memcmp(hd.bs, PBAP_TARGET, TARGET_SIZE) == 0 &&
-				os->server->services & OBEX_PBAP) {
-			os->target = PBAP_TARGET;
-			os->cmds = &pbap;
-			pbap_phonebook_context_create(os);
+		case OBEX_HDR_TARGET:
+			target = hd.bs;
+			target_size = hlen;
 			break;
 		}
+	}
 
+	os->service = obex_service_driver_find(os->server->drivers,
+						target, target_size,
+						who, who_size);
+	if (os->service == NULL) {
 		error("Connect attempt to a non-supported target");
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		return;
-	}
-
-	if (os->target == NULL) {
-		if (os->server->services & OBEX_OPP) {
-			register_transfer(os->cid, os);
-			/* OPP doesn't contains target or connection id. */
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
-		} else {
-			error("Object Push connect attempt to a non-OPP server");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		}
 
 		return;
 	}
 
-	register_session(cid, os);
-	emit_session_created(cid);
+	debug("Selected driver: %s", os->service->name);
 
-	/* Append received UUID in WHO header */
-	hd.bs = os->target;
-	OBEX_ObjectAddHeader(obex, obj,
-			OBEX_HDR_WHO, hd, TARGET_SIZE,
-			OBEX_FL_FIT_ONE_PACKET);
-	hd.bq4 = cid;
-	OBEX_ObjectAddHeader(obex, obj,
-			OBEX_HDR_CONNECTION, hd, 4,
-			OBEX_FL_FIT_ONE_PACKET);
-
-	OBEX_ObjectSetRsp(obj, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
+	if (os->service->connect)
+		os->service->connect(obex, obj);
 }
 
 static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
@@ -291,7 +241,7 @@ static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
 	os = OBEX_GetUserData(obex);
 
 	/* Object Push doesn't provide a connection id. */
-	if (os->target == NULL)
+	if (os->service->service == OBEX_OPP)
 		return TRUE;
 
 	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
@@ -316,10 +266,10 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 	guint hlen;
 	guint8 hi;
 
-	if (!os->cmds) {
+	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
-	} else if (!os->cmds->get) {
+	} else if (!os->service->get) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -327,21 +277,6 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 
 	g_return_if_fail(chk_cid(obex, obj, os->cid));
 
-	if (os->type) {
-		g_free(os->type);
-		os->type = NULL;
-	}
-
-	if (os->name) {
-		g_free(os->name);
-		os->name = NULL;
-	}
-
-	if (os->buf) {
-		g_free(os->buf);
-		os->buf = NULL;
-	}
-
 	while (OBEX_ObjectGetNextHeader(obex, obj, &hi, &hd, &hlen)) {
 		switch (hi) {
 		case OBEX_HDR_NAME:
@@ -379,11 +314,22 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 
 			os->type = g_strndup((const gchar *) hd.bs, hlen);
 			debug("OBEX_HDR_TYPE: %s", os->type);
+			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
 			break;
 		}
 	}
 
-	os->cmds->get(obex, obj);
+	if (!os->driver) {
+		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
+		if (!os->driver) {
+			error("No driver found");
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+					OBEX_RSP_NOT_IMPLEMENTED);
+			return;
+		}
+	}
+
+	os->service->get(obex, obj);
 }
 
 static void cmd_setpath(struct obex_session *os,
@@ -393,10 +339,10 @@ static void cmd_setpath(struct obex_session *os,
 	guint32 hlen;
 	guint8 hi;
 
-	if (!os->cmds) {
+	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
-	} else if (!os->cmds->setpath) {
+	} else if (!os->service->setpath) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -432,36 +378,32 @@ static void cmd_setpath(struct obex_session *os,
 		break;
 	}
 
-	os->cmds->setpath(obex, obj);
+	os->service->setpath(obex, obj);
 }
 
-int os_prepare_get(struct obex_session *os, gchar *file, guint32 *size)
+int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
 {
-	gint fd, err;
-	struct stat stats;
+	gint err;
+	gpointer object;
 
-	fd = open(file, O_RDONLY);
-	if (fd < 0)
+	object = os->driver->open(filename, O_RDONLY, 0, size);
+	if (object == NULL) {
+		err = -errno;
+		error("open(%s): %s (%d)", filename, strerror(errno), errno);
 		goto fail;
+	}
 
-	if (fstat(fd, &stats))
-		goto fail;
-
-	os->fd = fd;
+	os->object = object;
 	os->offset = 0;
 
-	if (stats.st_size > 0)
+	if (*size > 0)
 		os->buf = g_malloc0(os->tx_mtu);
 
-	*size = stats.st_size;
-
 	return 0;
 
 fail:
-	err = -errno;
-	if (fd >= 0)
-		close(fd);
-
+	if (object)
+		os->driver->close(object);
 	return err;
 }
 
@@ -472,14 +414,14 @@ static gint obex_write_stream(struct obex_session *os,
 	gint32 len;
 	guint8 *ptr;
 
-	debug("obex_write_stream: name=%s type=%s tx_mtu=%d fd=%d",
+	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
 		os->name ? os->name : "", os->type ? os->type : "",
-		os->tx_mtu, os->fd);
+		os->tx_mtu, os->object);
 
 	if (os->aborted)
 		return -EPERM;
 
-	if (os->fd < 0) {
+	if (os->object == NULL) {
 		if (os->buf == NULL && os->finished == FALSE)
 			return -EIO;
 
@@ -488,7 +430,7 @@ static gint obex_write_stream(struct obex_session *os,
 		goto add_header;
 	}
 
-	len = read(os->fd, os->buf, os->tx_mtu);
+	len = os->driver->read(os->object, os->buf, os->tx_mtu);
 	if (len < 0) {
 		gint err = errno;
 		error("read(): %s (%d)", strerror(err), err);
@@ -516,10 +458,6 @@ add_header:
 	OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, len,
 				OBEX_FL_STREAM_DATA);
 
-	if (!memcmp(os->target, PBAP_TARGET, TARGET_SIZE))
-		if (os->offset == os->size && os->finished == FALSE)
-			OBEX_SuspendRequest(obex, obj);
-
 	return len;
 }
 
@@ -530,8 +468,9 @@ gint os_prepare_put(struct obex_session *os)
 
 	path = g_build_filename(os->current_folder, os->name, NULL);
 
-	os->fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600);
-	if (os->fd < 0) {
+	os->object = os->driver->open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600,
+					(size_t *) &os->size);
+	if (os->object == NULL) {
 		error("open(%s): %s (%d)", path, strerror(errno), errno);
 		g_free(path);
 		return -EPERM;
@@ -539,9 +478,6 @@ gint os_prepare_put(struct obex_session *os)
 
 	g_free(path);
 
-	if (os->target == NULL)
-		emit_transfer_started(os->cid);
-
 	if (!os->buf) {
 		debug("PUT request checked, no buffered data");
 		return 0;
@@ -551,7 +487,7 @@ gint os_prepare_put(struct obex_session *os)
 	while (len < os->offset) {
 		gint w;
 
-		w = write(os->fd, os->buf + len, os->offset - len);
+		w = os->driver->write(os->object, os->buf + len, os->offset - len);
 		if (w < 0) {
 			gint err = errno;
 			error("write(%s): %s (%d)", path, strerror(errno),
@@ -593,15 +529,10 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 		return -EIO;
 	}
 
-	if (os->fd < 0 && size > 0) {
-		if (os->buf) {
-			error("Got more data but there is still a pending buffer");
-			return -EIO;
-		}
-
-		os->buf = g_malloc0(os->rx_mtu);
-		memcpy(os->buf, buffer, size);
-		os->offset = size;
+	if (os->object == NULL && size > 0) {
+		os->buf = g_realloc(os->buf, os->offset + size);
+		memcpy(os->buf + os->offset, buffer, size);
+		os->offset += size;
 
 		debug("Stored %u bytes into temporary buffer", size);
 
@@ -611,7 +542,7 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 	while (len < size) {
 		gint w;
 
-		w = write(os->fd, buffer + len, size - len);
+		w = os->driver->write(os->object, buffer + len, size - len);
 		if (w < 0) {
 			gint err = errno;
 			if (err == EINTR)
@@ -631,11 +562,9 @@ static gint obex_read_stream(struct obex_session *os, obex_t *obex,
 static gboolean check_put(obex_t *obex, obex_object_t *obj)
 {
 	struct obex_session *os;
-	struct statvfs buf;
 	obex_headerdata_t hd;
 	guint hlen;
 	guint8 hi;
-	long long unsigned int free;
 	int ret;
 
 	os = OBEX_GetUserData(obex);
@@ -686,6 +615,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 			os->type = g_strndup((const gchar *) hd.bs, hlen);
 			debug("OBEX_HDR_TYPE: %s", os->type);
+			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
 			break;
 
 		case OBEX_HDR_BODY:
@@ -705,18 +635,20 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 	OBEX_ObjectReParseHeaders(obex, obj);
 
-	if (!os->name) {
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		g_free(os->type);
-		os->type = NULL;
-		return FALSE;
+	if (!os->driver) {
+		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
+		if (!os->driver) {
+			error("No driver found");
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+					OBEX_RSP_NOT_IMPLEMENTED);
+			return FALSE;
+		}
 	}
 
-	if (!os->cmds || !os->cmds->chkput)
+	if (!os->service || !os->service->chkput)
 		goto done;
 
-	ret = os->cmds->chkput(obex, obj);
+	ret = os->service->chkput(obex, obj);
 	switch (ret) {
 	case 0:
 		break;
@@ -739,21 +671,6 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		goto done;
 	}
 
-	if (fstatvfs(os->fd, &buf) < 0) {
-		int err = errno;
-		error("fstatvfs(): %s(%d)", strerror(err), err);
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		return FALSE;
-	}
-
-	free = buf.f_bsize * buf.f_bavail;
-	debug("Free space in disk: %llu", free);
-	if ((guint64) os->size > free) {
-		debug("Free disk space not available");
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		return FALSE;
-	}
-
 done:
 	os->checked = TRUE;
 
@@ -762,10 +679,10 @@ done:
 
 static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
-	if (!os->cmds) {
+	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
-	} else if (!os->cmds->put) {
+	} else if (!os->service->put) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 				OBEX_RSP_NOT_IMPLEMENTED);
 		return;
@@ -778,7 +695,7 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			return;
 	}
 
-	os->cmds->put(obex, obj);
+	os->service->put(obex, obj);
 }
 
 static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
@@ -792,14 +709,13 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
-		/* Just emit progress for Object Push */
-		if (os->target == NULL)
-			emit_transfer_progress(os->cid, os->size, os->offset);
+		if (os->service->progress)
+			os->service->progress(obex, obj);
 		break;
 	case OBEX_EV_ABORT:
 		os->aborted = TRUE;
-		if (os->target == NULL)
-			emit_transfer_completed(os->cid, FALSE);
+		if (os->service->reset)
+			os->service->reset(obex);
 		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
@@ -810,9 +726,10 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			break;
 		case OBEX_CMD_PUT:
 		case OBEX_CMD_GET:
+		case OBEX_CMD_SETPATH:
 			os_session_mark_aborted(os);
-			if (os->target == NULL)
-				emit_transfer_completed(os->cid, !os->aborted);
+			if (os->service->reset)
+				os->service->reset(obex);
 			os_reset_session(os);
 			break;
 		default:
@@ -841,7 +758,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_REQCHECK:
 		switch (cmd) {
 		case OBEX_CMD_PUT:
-			if (os->cmds && os->cmds->put)
+			if (os->service && os->service->put)
 				check_put(obex, obj);
 			break;
 		default:
@@ -905,7 +822,6 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 
 void server_free(struct server *server)
 {
-	g_free(server->name);
 	g_free(server->folder);
 	g_free(server->capability);
 	g_free(server->devnode);
@@ -919,16 +835,8 @@ static void obex_handle_destroy(gpointer user_data)
 
 	os = OBEX_GetUserData(obex);
 
-	if (os->target == NULL) {
-		/* Got an error during a transfer. */
-		if (os->fd >= 0)
-			emit_transfer_completed(os->cid, os->offset == os->size);
-
-		unregister_transfer(os->cid);
-	} else {
-		emit_session_removed(os->cid);
-		unregister_session(os->cid);
-	}
+	if (os->service && os->service->disconnect)
+		os->service->disconnect(obex);
 
 	obex_session_free(os);
 
@@ -938,8 +846,17 @@ static void obex_handle_destroy(gpointer user_data)
 static gboolean tty_reinit(gpointer data)
 {
 	struct server *server = data;
+	GSList *l;
+	guint services = 0;
 
-	tty_init(server->services, server->folder, server->capability, server->devnode);
+	for (l = server->drivers; l; l = l->next) {
+		struct obex_service_driver *driver = l->data;
+
+		services |= driver->service;
+	}
+
+	tty_init(services, server->folder, server->capability,
+			server->symlinks, server->devnode);
 
 	server_free(server);
 
@@ -986,16 +903,12 @@ gint obex_session_start(GIOChannel *io, struct server *server)
 
 	os = g_new0(struct obex_session, 1);
 
-	os->target = NULL;
-
-	if (server->services & OBEX_OPP)
-		os->cmds = &opp;
-
+	os->service = obex_service_driver_find(server->drivers, NULL, 0,
+						NULL, 0);
 	os->current_folder = g_strdup(server->folder);
 	os->server = server;
 	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
 	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
-	os->fd = -1;
 	os->size = OBJECT_SIZE_DELETE;
 
 	obex = OBEX_Init(OBEX_TRANS_FD, obex_event, 0);
@@ -1041,3 +954,17 @@ gint obex_tty_session_stop(void)
 
 	return 0;
 }
+
+struct obex_session *obex_get_session(gpointer object)
+{
+	GSList *l;
+
+	for (l = sessions; l; l = l->next) {
+		struct obex_session *os = l->data;
+
+		if (os->object == object)
+			return os;
+	}
+
+	return NULL;
+}
